// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::{MOpTrait, PRegNum, UnitOutRegNum, COMMON_MOP_SRC_LEN},
    register::PRegValue,
    unit::{UnitCancelInput, UnitOutput, UnitOutputWrite},
    util::tree_reduce::tree_reduce,
};
use fayalite::{module::wire_with_loc, prelude::*, ty::StaticType, util::ready_valid::ReadyValid};
use std::marker::PhantomData;

#[hdl]
pub struct UnitForwardingInfo<UnitNumWidth: Size, OutRegNumWidth: Size, UnitCount: Size> {
    pub unit_output_writes: ArrayType<HdlOption<UnitOutputWrite<OutRegNumWidth>>, UnitCount>,
    pub _phantom: PhantomData<UnitNumWidth>,
}

#[hdl]
pub struct UnitToRegAlloc<
    MOp: Type,
    ExtraOut: Type,
    UnitNumWidth: Size,
    OutRegNumWidth: Size,
    UnitCount: Size,
> {
    #[hdl(flip)]
    pub unit_forwarding_info: UnitForwardingInfo<UnitNumWidth, OutRegNumWidth, UnitCount>,
    #[hdl(flip)]
    pub input_insn: ReadyValid<MOp>,
    #[hdl(flip)]
    pub cancel_input: HdlOption<UnitCancelInput<OutRegNumWidth>>,
    pub output: HdlOption<UnitOutput<OutRegNumWidth, ExtraOut>>,
}

impl<MOp: Type, ExtraOut: Type, UnitNumWidth: Size, OutRegNumWidth: Size, UnitCount: Size>
    UnitToRegAlloc<MOp, ExtraOut, UnitNumWidth, OutRegNumWidth, UnitCount>
{
    pub fn mop_ty(self) -> MOp {
        self.input_insn.data.HdlSome
    }
    pub fn extra_out_ty(self) -> ExtraOut {
        self.output.HdlSome.extra_out_ty()
    }
}

#[hdl]
pub struct ExecuteStart<MOp: Type + MOpTrait<DestReg = UnitOutRegNum<DynSize>>> {
    pub mop: MOp,
    pub src_values: Array<PRegValue, { COMMON_MOP_SRC_LEN }>,
}

#[hdl]
pub struct ExecuteEnd<OutRegNumWidth: Size, ExtraOut> {
    pub unit_output: UnitOutput<OutRegNumWidth, ExtraOut>,
}

#[hdl]
enum InFlightOpState {
    Ready,
    Running,
    CanceledAndRunning,
}

impl InFlightOpState {
    fn ready_next_state(canceling: bool, starting: bool, ending: bool) -> Expr<HdlOption<Self>> {
        match (canceling, starting, ending) {
            (false, false, _) => HdlSome(InFlightOpState.Ready()),
            (false, true, false) => HdlSome(InFlightOpState.Running()),
            (false, true, true) => HdlNone(),
            (true, false, _) => HdlNone(),
            (true, true, false) => HdlSome(InFlightOpState.CanceledAndRunning()),
            (true, true, true) => HdlNone(),
        }
    }
    fn running_next_state(canceling: bool, _starting: bool, ending: bool) -> Expr<HdlOption<Self>> {
        match (canceling, ending) {
            (false, false) => HdlSome(InFlightOpState.Running()),
            (false, true) => HdlNone(),
            (true, false) => HdlSome(InFlightOpState.CanceledAndRunning()),
            (true, true) => HdlNone(),
        }
    }
    fn canceled_and_running_next_state(
        _canceling: bool,
        _starting: bool,
        ending: bool,
    ) -> Expr<HdlOption<Self>> {
        if ending {
            HdlNone()
        } else {
            HdlSome(InFlightOpState.CanceledAndRunning())
        }
    }
    /// FIXME: this is working around #[hdl] match not supporting matching values inside structs yet
    #[hdl]
    fn connect_next_state(
        canceling: Expr<Bool>,
        starting: Expr<Bool>,
        ending: Expr<Bool>,
        next_state_fn: fn(canceling: bool, starting: bool, ending: bool) -> Expr<HdlOption<Self>>,
        next_state: Expr<HdlOption<Self>>,
    ) {
        #[hdl]
        fn recurse<const N: usize>(
            exprs: &[Expr<Bool>; N],
            bools: &mut [bool; N],
            f: &mut impl FnMut(&[bool; N]),
            arg_index: usize,
        ) {
            if arg_index < N {
                #[hdl]
                if exprs[arg_index] {
                    bools[arg_index] = true;
                    recurse(exprs, bools, f, arg_index + 1);
                } else {
                    bools[arg_index] = false;
                    recurse(exprs, bools, f, arg_index + 1);
                }
            } else {
                f(bools);
            }
        }
        recurse(
            &[canceling, starting, ending],
            &mut [false; 3],
            &mut |&[canceling, starting, ending]| {
                connect(next_state, next_state_fn(canceling, starting, ending))
            },
            0,
        );
    }
}

#[hdl]
struct InFlightOp<MOp: Type> {
    state: InFlightOpState,
    mop: MOp,
    src_ready_flags: Array<Bool, { COMMON_MOP_SRC_LEN }>,
}

#[hdl]
struct InFlightOpsSummary<OpIndexWidth: Size> {
    empty_op_index: HdlOption<UIntType<OpIndexWidth>>,
    ready_op_index: HdlOption<UIntType<OpIndexWidth>>,
}

impl<OpIndexWidth: Size> InFlightOpsSummary<OpIndexWidth> {
    #[hdl]
    fn new<MOp: Type>(
        op_index: usize,
        op_index_ty: UIntType<OpIndexWidth>,
        in_flight_op: impl ToExpr<Type = HdlOption<InFlightOp<MOp>>>,
    ) -> Expr<Self> {
        let empty_op_index = wire_with_loc(
            &format!("empty_op_index_{op_index}"),
            SourceLocation::caller(),
            HdlOption[op_index_ty],
        );
        connect(empty_op_index, HdlOption[op_index_ty].HdlNone());
        let ready_op_index = wire_with_loc(
            &format!("ready_op_index_{op_index}"),
            SourceLocation::caller(),
            HdlOption[op_index_ty],
        );
        connect(ready_op_index, HdlOption[op_index_ty].HdlNone());
        #[hdl]
        if let HdlSome(in_flight_op) = in_flight_op {
            #[hdl]
            let InFlightOp::<_> {
                state,
                mop: _,
                src_ready_flags,
            } = in_flight_op;
            connect(ready_op_index, HdlOption[op_index_ty].HdlNone());
            #[hdl]
            match state {
                InFlightOpState::Ready =>
                {
                    #[hdl]
                    if src_ready_flags.cmp_eq([true; COMMON_MOP_SRC_LEN]) {
                        connect(ready_op_index, HdlSome(op_index.cast_to(op_index_ty)));
                    }
                }
                InFlightOpState::CanceledAndRunning | InFlightOpState::Running => {}
            }
        } else {
            connect(empty_op_index, HdlSome(op_index.cast_to(op_index_ty)));
        }
        #[hdl]
        InFlightOpsSummary::<_> {
            empty_op_index,
            ready_op_index,
        }
    }
    #[hdl]
    fn combine(l: impl ToExpr<Type = Self>, r: impl ToExpr<Type = Self>) -> Expr<Self> {
        let l = l.to_expr();
        let r = r.to_expr();
        #[hdl]
        InFlightOpsSummary::<_> {
            empty_op_index: HdlOption::or(l.empty_op_index, r.empty_op_index),
            ready_op_index: HdlOption::or(l.ready_op_index, r.ready_op_index),
        }
    }
}

impl InFlightOpsSummary<DynSize> {
    fn summarize<MOp: Type, MaxInFlight: Size>(
        in_flight_ops: impl ToExpr<Type = ArrayType<HdlOption<InFlightOp<MOp>>, MaxInFlight>>,
    ) -> Expr<Self> {
        let in_flight_ops = in_flight_ops.to_expr();
        let max_in_flight = Expr::ty(in_flight_ops).len();
        let index_range = 0..max_in_flight;
        let index_ty = UInt::range(index_range.clone());
        tree_reduce(
            index_range.map(|i| Self::new(i, index_ty, in_flight_ops[i])),
            Self::combine,
        )
        .expect("in_flight_ops is known to have len > 0")
    }
}

#[hdl_module]
pub fn unit_base<
    MOp: Type + MOpTrait<DestReg = UnitOutRegNum<DynSize>, SrcRegWidth = DynSize>,
    ExtraOut: Type,
>(
    config: &CpuConfig,
    unit_index: usize,
    mop_ty: MOp,
    extra_out_ty: ExtraOut,
) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let unit_to_reg_alloc: UnitToRegAlloc<MOp, ExtraOut, DynSize, DynSize, DynSize> =
        m.output(config.unit_to_reg_alloc(mop_ty, extra_out_ty));
    #[hdl]
    let execute_start: ReadyValid<ExecuteStart<MOp>> = m.output(ReadyValid[ExecuteStart[mop_ty]]);
    #[hdl]
    let execute_end: HdlOption<ExecuteEnd<DynSize, ExtraOut>> =
        m.input(HdlOption[ExecuteEnd[config.out_reg_num_width][extra_out_ty]]);

    connect(execute_start.data, Expr::ty(execute_start).data.HdlNone());

    let max_in_flight = config.unit_max_in_flight(unit_index).get();
    let in_flight_op_ty = InFlightOp[mop_ty];
    #[hdl]
    let in_flight_ops = reg_builder()
        .clock_domain(cd)
        .reset(repeat(HdlOption[in_flight_op_ty].HdlNone(), max_in_flight));

    let in_flight_ops_summary_value = InFlightOpsSummary::summarize(in_flight_ops);
    #[hdl]
    let in_flight_ops_summary = wire(Expr::ty(in_flight_ops_summary_value));
    connect(in_flight_ops_summary, in_flight_ops_summary_value);

    connect(
        unit_to_reg_alloc.input_insn.ready,
        HdlOption::is_some(in_flight_ops_summary.empty_op_index),
    );
    // TODO: connect(execute_start.data, <read_regs>(in_flight_ops_summary.ready_op_index));

    connect(
        unit_to_reg_alloc.output,
        Expr::ty(unit_to_reg_alloc.output).HdlNone(),
    ); // TODO: finish

    #[hdl]
    let input_in_flight_op = wire(HdlOption[in_flight_op_ty]);
    connect(input_in_flight_op, HdlOption[in_flight_op_ty].HdlNone());
    #[hdl]
    if let HdlSome(mop) = ReadyValid::firing_data(unit_to_reg_alloc.input_insn) {
        #[hdl]
        let input_mop_src_regs = wire(mop_ty.src_regs_ty());
        connect(
            input_mop_src_regs,
            repeat(config.p_reg_num().const_zero().cast_to_bits(), ConstUsize),
        );
        MOp::connect_src_regs(mop, input_mop_src_regs);
        let src_ready_flags = wire_with_loc(
            "input_in_flight_op_src_ready_flags",
            SourceLocation::caller(),
            StaticType::TYPE,
        );
        for src_index in 0..COMMON_MOP_SRC_LEN {
            connect(
                src_ready_flags[src_index],
                config
                    .p_reg_num()
                    .const_zero()
                    .cast_to_bits()
                    .cmp_eq(input_mop_src_regs[src_index]),
            );
        }
        #[hdl]
        if unit_to_reg_alloc.cancel_input.cmp_ne(HdlSome(
            #[hdl]
            UnitCancelInput::<_> {
                which: MOp::dest_reg(mop),
            },
        )) {
            connect(
                input_in_flight_op,
                HdlSome(
                    #[hdl]
                    InFlightOp::<_> {
                        state: InFlightOpState.Ready(),
                        mop,
                        src_ready_flags,
                    },
                ),
            );
        }
        #[hdl]
        if let HdlSome(empty_op_index) = in_flight_ops_summary.empty_op_index {
            connect(in_flight_ops[empty_op_index], input_in_flight_op);
        }
    }

    #[hdl]
    let in_flight_op_next_state = wire(Array[HdlOption[InFlightOpState]][max_in_flight]);
    #[hdl]
    let in_flight_op_next_src_ready_flags =
        wire(Array[in_flight_op_ty.src_ready_flags][max_in_flight]);
    #[hdl]
    let in_flight_op_canceling = wire(Array[Bool][max_in_flight]);
    #[hdl]
    let in_flight_op_execute_starting = wire(Array[Bool][max_in_flight]);
    #[hdl]
    let in_flight_op_execute_ending = wire(Array[Bool][max_in_flight]);
    for in_flight_op_index in 0..max_in_flight {
        connect(
            in_flight_op_next_src_ready_flags[in_flight_op_index],
            [false; COMMON_MOP_SRC_LEN],
        );
        connect(in_flight_op_canceling[in_flight_op_index], false);
        connect(in_flight_op_execute_starting[in_flight_op_index], false);
        connect(in_flight_op_execute_ending[in_flight_op_index], false);
        #[hdl]
        if let HdlSome(in_flight_op) = in_flight_ops[in_flight_op_index] {
            #[hdl]
            let InFlightOp::<_> {
                state,
                mop,
                src_ready_flags,
            } = in_flight_op;
            let which = MOp::dest_reg(mop);
            let src_regs = wire_with_loc(
                &format!("in_flight_op_src_regs_{in_flight_op_index}"),
                SourceLocation::caller(),
                mop_ty.src_regs_ty(),
            );
            connect(
                src_regs,
                repeat(config.p_reg_num().const_zero().cast_to_bits(), ConstUsize),
            );
            MOp::connect_src_regs(mop, src_regs);

            connect(
                in_flight_op_next_src_ready_flags[in_flight_op_index],
                src_ready_flags,
            );
            let unit_output_writes = unit_to_reg_alloc.unit_forwarding_info.unit_output_writes;
            for unit_index in 0..Expr::ty(unit_output_writes).len() {
                #[hdl]
                if let HdlSome(unit_output_write) = unit_output_writes[unit_index] {
                    #[hdl]
                    let UnitOutputWrite::<_> {
                        which: unit_out_reg,
                        value: _,
                    } = unit_output_write;
                    let p_reg_num = #[hdl]
                    PRegNum::<_, _> {
                        unit_num: config.unit_num().from_index(unit_index),
                        unit_out_reg,
                    };
                    for src_index in 0..COMMON_MOP_SRC_LEN {
                        #[hdl]
                        if p_reg_num.cast_to_bits().cmp_eq(src_regs[src_index]) {
                            connect(
                                in_flight_op_next_src_ready_flags[in_flight_op_index][src_index],
                                true,
                            );
                        }
                    }
                }
            }

            connect(
                in_flight_op_canceling[in_flight_op_index],
                unit_to_reg_alloc.cancel_input.cmp_eq(HdlSome(
                    #[hdl]
                    UnitCancelInput::<_> { which },
                )),
            );

            #[hdl]
            if let HdlSome(execute_end) = execute_end {
                #[hdl]
                let ExecuteEnd::<_, _> { unit_output } = execute_end;
                #[hdl]
                if which.cmp_eq(unit_output.which) {
                    connect(in_flight_op_execute_ending[in_flight_op_index], true);
                }
            }
            #[hdl]
            if let HdlSome(execute_start) = ReadyValid::firing_data(execute_start) {
                #[hdl]
                if which.cmp_eq(MOp::dest_reg(execute_start.mop)) {
                    connect(in_flight_op_execute_starting[in_flight_op_index], true);
                }
            }
            let connect_next_state = |f| {
                InFlightOpState::connect_next_state(
                    in_flight_op_canceling[in_flight_op_index],
                    in_flight_op_execute_starting[in_flight_op_index],
                    in_flight_op_execute_ending[in_flight_op_index],
                    f,
                    in_flight_op_next_state[in_flight_op_index],
                );
            };
            #[hdl]
            match state {
                InFlightOpState::Ready => connect_next_state(InFlightOpState::ready_next_state),
                InFlightOpState::Running => connect_next_state(InFlightOpState::running_next_state),
                InFlightOpState::CanceledAndRunning => {
                    connect_next_state(InFlightOpState::canceled_and_running_next_state);
                }
            }
            #[hdl]
            if let HdlSome(state) = in_flight_op_next_state[in_flight_op_index] {
                connect(
                    in_flight_ops[in_flight_op_index],
                    HdlSome(
                        #[hdl]
                        InFlightOp::<_> {
                            state,
                            mop,
                            src_ready_flags,
                        },
                    ),
                );
            } else {
                connect(
                    in_flight_ops[in_flight_op_index],
                    HdlOption[in_flight_op_ty].HdlNone(),
                );
            }
        } else {
            connect(in_flight_op_next_state[in_flight_op_index], HdlNone());
        }
    }
}
